@@ -70,6 +70,13 @@
 #include "compat-bsd.h"
 #include "compat-ssl.h"
 
+/* GCC function attributes */
+#ifdef HAVE_FUNC_ATTRIBUTE_FORMAT
+#define PRINTF_FUNC(argn) __attribute__((format(__printf__, argn, argn+1)))
+#else
+#define PRINTF_FUNC(argn)
+#endif
+
 /* Internationalization */
 #ifdef ENABLE_NLS
 #define _(x)			gettext(x)
@@ -94,6 +101,7 @@ typedef struct {
 typedef message_t url_t;
 typedef message_t if_t;
 
+#include "abuf.h"
 #include "conf.h"
 #include "tcp.h"
 #include "ftp.h"
