@@ -273,16 +273,21 @@ conn_init(conn_t *conn)
 		if (!http_connect(conn->http, conn->proto, proxy, conn->host,
 				  conn->port, conn->user, conn->pass,
 				  conn->conf->io_timeout)) {
-			conn->message = conn->http->headers;
+			conn->message = conn->http->headers->p;
 			conn_disconnect(conn);
 			return 0;
 		}
-		conn->message = conn->http->headers;
+		conn->message = conn->http->headers->p;
 		conn->tcp = &conn->http->tcp;
 	}
 	return 1;
 }
 
+/**
+ * Setup the connection.
+ *
+ * Must be called with the conn_t lock held.
+ */
 int
 conn_setup(conn_t *conn)
 {
@@ -310,6 +315,8 @@ conn_setup(conn_t *conn)
 		conn->http->firstbyte =
 			conn->supported ? conn->currentbyte : -1;
 		conn->http->lastbyte = conn->lastbyte;
+
+		abuf_setup(conn->http->request, 2048);
 		http_get(conn->http, s);
 		for (i = 0; i < conn->conf->add_header_count; i++)
 			http_addheader(conn->http, "%s",
@@ -326,6 +333,7 @@ conn_exec(conn_t *conn)
 			return 0;
 		return ftp_wait(conn->ftp) / 100 == 1;
 	} else {
+		abuf_setup(conn->http->headers, 1024);
 		if (!http_exec(conn->http))
 			return 0;
 		return conn->http->status / 100 == 2;
@@ -368,89 +376,90 @@ conn_info(conn_t *conn)
 	/* It's all a bit messed up.. But it works. */
 	if (PROTO_IS_FTP(conn->proto) && !conn->proxy) {
 		return conn_info_ftp(conn);
-	} else {
-		char s[1005];
-		long long int i = 0;
+	}
 
-		do {
-			const char *t;
+	char s[1005];
+	long long int i = 0;
 
-			conn->supported = true;
-			conn->currentbyte = 0;
-			if (!conn_setup(conn))
-				return 0;
-			conn_exec(conn);
-			conn_disconnect(conn);
+	do {
+		const char *t;
 
-			http_filename(conn->http, conn->output_filename);
-
-			/* Code 3xx == redirect */
-			if (conn->http->status / 100 != 3)
-				break;
-			if ((t = http_header(conn->http, "location:")) == NULL)
-				return 0;
-			sscanf(t, "%1000s", s);
-			if (s[0] == '/') {
-				snprintf(conn->http->headers,
-					 sizeof(conn->http->headers),
-					 "%s%s:%i%s",
-					 scheme_from_proto(conn->proto),
-					 conn->host, conn->port, s);
-				strlcpy(s, conn->http->headers, sizeof(s));
-			} else if (strstr(s, "://") == NULL) {
-				conn_url(conn->http->headers,
-					 sizeof(conn->http->headers), conn);
-				strlcat(conn->http->headers, s,
-					sizeof(conn->http->headers));
-				strlcpy(s, conn->http->headers, sizeof(s));
-			}
-
-			if (!conn_set(conn, s)) {
-				return 0;
-			}
-
-			/* check if the download has been redirected to FTP and
-			 * report it back to the caller */
-			if (PROTO_IS_FTP(conn->proto) && !conn->proxy) {
-				return -1;
-			}
-
-			if (++i >= conn->conf->max_redirect) {
-				fprintf(stderr, _("Too many redirects.\n"));
-				return 0;
-			}
-		} while (conn->http->status / 100 == 3);
-
-		/* Check for non-recoverable errors */
-		if (conn->http->status != 416 && conn->http->status / 100 != 2)
+		conn->supported = true;
+		conn->currentbyte = 0;
+		pthread_mutex_lock(&conn->lock);
+		if (!conn_setup(conn))
 			return 0;
+		pthread_mutex_unlock(&conn->lock);
+		conn_exec(conn);
+		conn_disconnect(conn);
 
-		conn->size = http_size_from_range(conn->http);
-		/* We assume partial requests are supported if a Content-Range
-		 * header is present.
-		 */
-		conn->supported = conn->http->status == 206 || conn->size > 0;
+		http_filename(conn->http, conn->output_filename);
 
-		if (conn->size <= 0) {
-			/* Sanity check */
-			switch (conn->http->status) {
-			case 200:
-			case 206:
-			case 416:
-				break;
-			default: /* unexpected */
-				return 0;
-			}
-
-			/* So we got an invalid or no size, fall back */
-			conn->supported = false;
-			conn->size = LLONG_MAX;
-		} else {
-			/* If Content-Length and Content-Range disagree, it's a
-			 * server bug; we take the larger and hope for the best.
-			 */
-			conn->size = max(conn->size, http_size(conn->http));
+		/* Code 3xx == redirect */
+		if (conn->http->status / 100 != 3)
+			break;
+		if ((t = http_header(conn->http, "location:")) == NULL)
+			return 0;
+		sscanf(t, "%1000s", s);
+		if (s[0] == '/') {
+			abuf_printf(conn->http->headers, "%s%s:%i%s",
+				    scheme_from_proto(conn->proto),
+				    conn->host, conn->port, s);
+			strlcpy(s, conn->http->headers->p, sizeof(s));
+		} else if (strstr(s, "://") == NULL) {
+			conn_url(conn->http->headers->p,
+				 conn->http->headers->len,
+				 conn);
+			strlcat(conn->http->headers->p, s,
+				conn->http->headers->len);
+			strlcpy(s, conn->http->headers->p, sizeof(s));
 		}
+
+		if (!conn_set(conn, s)) {
+			return 0;
+		}
+
+		/* check if the download has been redirected to FTP and
+		 * report it back to the caller */
+		if (PROTO_IS_FTP(conn->proto) && !conn->proxy) {
+			return -1;
+		}
+
+		if (++i >= conn->conf->max_redirect) {
+			fprintf(stderr, _("Too many redirects.\n"));
+			return 0;
+		}
+	} while (conn->http->status / 100 == 3);
+
+	/* Check for non-recoverable errors */
+	if (conn->http->status != 416 && conn->http->status / 100 != 2)
+		return 0;
+
+	conn->size = http_size_from_range(conn->http);
+	/* We assume partial requests are supported if a Content-Range
+	 * header is present.
+	 */
+	conn->supported = conn->http->status == 206 || conn->size > 0;
+
+	if (conn->size <= 0) {
+		/* Sanity check */
+		switch (conn->http->status) {
+		case 200:
+		case 206:
+		case 416:
+			break;
+		default: /* unexpected */
+			return 0;
+		}
+
+		/* So we got an invalid or no size, fall back */
+		conn->supported = false;
+		conn->size = LLONG_MAX;
+	} else {
+		/* If Content-Length and Content-Range disagree, it's a
+		 * server bug; we take the larger and hope for the best.
+		 */
+		conn->size = max(conn->size, http_size(conn->http));
 	}
 
 	return 1;
